
import java.math.BigInteger;
import java.util.Vector;
import java.lang.Exception;

/**
 * Diese Klasse implementiert einen Mix.
 * @author Sebastian Clauß
 * @version 1.0
 */


public class Mix {
  /**
   * Nummer des nächsten Schubes bzw. Anzahl bisheriger Schübe
   */
  
  private int batchNumber;
  /**
   * Dynamisches Feld, das alle bisherigen eingehenden Schübe enthält
   */
  
  private Vector<BigInteger[]> inBatches;
  /**
   * Dynamisches Feld, daß alle bisherigen ausgehenden Schübe enthält
   */
  
  private Vector<BigInteger[]> outBatches;
  /**
   * Anzahl der Nachrichten in einem Schub
   */
  
  public static int BATCH_SIZE = 32;
  /**
   * Anteil der Zufallsbits in einer Nachricht in Prozent
   */
  
  public static int RANDOM_AMOUNT = 20;
  /**
   * Länge einer Nachricht
   */
  
  private int messageLength;
  /**
   * Länge der Zufallsbitfolge in einer Nachricht
   */
  
  private int randomLength;

  /**
   * Der geheime RSA-Schlüssel des Mixes
   */
  
  private RSAPrivateKey mixPrivateKey;
  /**
   * der öffentliche RSA-Schlüssel des Mixes
   */
  
  private RSAPublicKey mixPublicKey;
  /**
   * Schlüssellänge des RSA-Schlüssels
   */
  
  private static int KEY_LENGTH = 512;

  /**
   * Anzahl der im aktuellen Schub bereits eingegangenen Nachrichten
   */
  
  private int numInMessages = 0;

  /**
   * Tools zum Behandeln von Nachrichten
   */
  
  private MessageTools messageTools;

  /**
   * Initialisierung des Mixes. Es werden die RSA-Schlüssel erzeugt, und
   * die Felder für Ein- und Ausgabenachrichten werden initialisiert.
   */
  
  public Mix(){
    System.out.print(Attack.messages.getString("MixGenKey"));
    do{
      RSAKeyGenerator gen = new RSAKeyGenerator(KEY_LENGTH);
      mixPrivateKey = gen.getPrivateKey();
      mixPublicKey = gen.getPublicKey();
    }while(mixPublicKey.n.compareTo(new BigInteger("1").shiftLeft(KEY_LENGTH-2)) < 0);
    System.out.println(Attack.messages.getString("MixGenKey2"));

    //System.out.println("Länge von n: " +Integer.toString(mixPublicKey.n.bitLength()));

    messageLength = KEY_LENGTH - 2;
    if (messageLength < 0) messageLength = 0;
    randomLength = (messageLength * RANDOM_AMOUNT) / 100;
    // Eingabeschubdatenstruktur für ersten Schub anlegen
    // create input data structure for the first batch
    inBatches = new Vector<BigInteger[]>(10,5);
    inBatches.add(new BigInteger[BATCH_SIZE]);
    outBatches = new Vector<BigInteger[]>(10,5);

    messageTools = new MessageTools(this);
  }

  /**
   * Bearbeiten eines Schubes im Mix. Falls der Schub noch nicht voll ist, wird er
   * zunächst mit zufälligen Nachrichten aufgefüllt. Die Nachrichten werden dann
   * umkodiert (entschlüsselt mit dem geheimen RSA-Schlüssel des Mixes) und umsortiert.
   */
  
  public void executeBatch(){
    // Schub auffüllen mit zufälligen Nachrichten und Duplikate erkennen
    // fill the batch with random messages and make out duplicates
    boolean batchNotFull = true;
    int i = 1;
    while (batchNotFull){
      try{
        batchNotFull
          = addInMessage(messageTools.generateRandomMessage());
      }catch(Exception e){
        System.out.println(Attack.messages.getString("MixMsgLen"));
        e.printStackTrace();
        System.exit(-1);
      }
      i++;
    }
    // umcodieren
    // code the messages
    BigInteger outBatch[] = new BigInteger[BATCH_SIZE];
    for(i = 0; i< BATCH_SIZE; i++){
      outBatch[i] = RSA.decrypt(mixPrivateKey, ((BigInteger[])inBatches.lastElement())[i]);
      // entfernen der Zufallsbits
      // remove the random bits
      outBatch[i] = outBatch[i].xor(outBatch[i].shiftRight(messageLength-randomLength)
                                               .shiftLeft(messageLength-randomLength));
    }
    // umsortieren
    // sort
    for(i = 0; i < BATCH_SIZE; i++){
      for (int j = i+1; j < BATCH_SIZE; j++){
        if(outBatch[i].compareTo(outBatch[j]) > 0){
          BigInteger a = outBatch[i];
          outBatch[i] = outBatch[j];
          outBatch[j] = a;
        }
      }
    }

    outBatches.add(outBatch);
    // Eingabeschubdatenstruktur für nächsten Schub anlegen
    // create input data structure for the next batch
    batchNumber++;
    inBatches.add(new BigInteger[BATCH_SIZE]);
    numInMessages = 0;
  }


  /**
   * Fügt eine Eingabenachricht zum aktuellen Schub hinzu, wenn darin noch Platz ist.
   * @param m Die Nachricht
   * @return true, wenn die Nachricht noch hineinpaßt, sonst false
   * @exception Exception wenn die Nachricht zu groß ist (m >= n oder m <= -n). n ist der
   *                      Modulus des RSA-Schlüssels des Mixes
   */
  
  public boolean addInMessage(BigInteger m)throws Exception{
    if((m.compareTo(mixPublicKey.n) >= 0) || (m.compareTo(mixPublicKey.n.negate()) <= 0)){
      throw new Exception(Attack.messages.getString("MixMsgLen2"));
    }
    if(m.compareTo(BigInteger.ZERO) < 0) m = m.add(mixPublicKey.n);

    // erkenne Wiederholungen
    // detect replays
    for(int i = 0; i < inBatches.size(); i++) {
    	BigInteger[] batch = inBatches.elementAt(i);
    	if(batch != null) {
    		if(i == batchNumber) {
    			for (int j = 0; j < numInMessages; j++) {
    				if(m.compareTo(batch[j]) == 0) {
    					// DUPLIKAT: Verwerfe die Nachricht, aber gib true zurück (Eingabe korrekt, aber nicht benutzbar)
    					// DUPLICATE: Do not add the message, but return true (input correct, but not usable)
    					return true;
    				}
    			}
    		} else {
    			for (int j = 0; j < BATCH_SIZE; j++) {
    				if(m.compareTo(batch[j]) == 0) {
    					// DUPLIKAT: Verwerfe die Nachricht, aber gib true zurück (Eingabe korrekt, aber nicht benutzbar)
    					// DUPLICATE: Do not add the message, but return true (input correct, but not allowed)
    					return true;
    				}
    			}
    		}
    	}
    }
    
    if (numInMessages < BATCH_SIZE){
      ((BigInteger[])inBatches.lastElement())[numInMessages] = m;
      numInMessages++;
      return true;
    }else{
      return false;
    }
  }

  /**
   * Gibt die maximale Länge einer Nachricht in Bit zurück.
   * @return die maximale Länge einer Nachricht in Bit
   */
  
  public int getMessageLength(){
    return messageLength;
  }

  /**
   * Gibt die Länge der Zufallsbitfolge in einer Nachricht an.
   * @return die Länge der Zufallsbitfolge in einer Nachricht
   */
  
  public int getRandomLength(){
    return randomLength;
  }

  /**
   * Gibt den öffentlichen RSA-Schlüssel des Mixes zurück.
   * @return den öffentlichen RSA-Schlüssel des Mixes
   */
  
  public RSAPublicKey getPublicKey(){
    return mixPublicKey;
  }

  /**
   * Gibt die Klasse mit den Nachrichtenbearbeitungsfunktionen
   * für diesen Mix zurück. Die Klasse ist bereits vorinitialisiert
   * mit den Parametern des Mixes (Nachrichtenlänge, Anzahl Zufallsbits
   * pro Nachricht, öffentlicher Schlüssel des Mixes).
   * @return die MessageTools abgestimmt auf diesen Mix
   */
  
  public MessageTools getMessageTools(){
    return messageTools;
  }

  /**
   * Gibt die aktuelle Schubnummer zurück.
   * @return die aktuelle Schubnummer
   */
  
  public int getBatchNumber(){
    return batchNumber;
  }

  /**
   * Gibt ein BigInteger-Feld mit den Eingabenachrichten des durch "index"
   * bezeichneten Schubes zurück. Ein Schub enthält "BATCH_SIZE" Nachrichten.
   * Die Positionen der Nachrichten im Schub sind konstant.
   * @param index der Index des Schubes, der zurückgegeben werden soll
   * @return null, falls "index" außerhalb der bereits bearbeiteten Schübe liegt,
   *         sonst ein BigInteger-Feld mit den Eingabenachrichten des durch "index" bezeichneten
   *         Schubes
   */
  
  public BigInteger[] getInBatch(int index){
    if((index >= 0) && (index < batchNumber)){
    	// clone the batch before outputting it 
    	BigInteger[] result = new BigInteger[BATCH_SIZE];
    	for(int i = 0; i < BATCH_SIZE; i++) {
    		result[i] = inBatches.elementAt(index)[i];
    	}
    	return result;
    }
    return null;
  }

  /**
   * Gibt ein BigInteger-Feld mit den Ausgabenachrichten des durch "index"
   * bezeichneten Schubes zurück. Ein Schub enthält "BATCH_SIZE" Nachrichten.
   * Die Positionen der Nachrichten im Schub sind konstant.
   * @param index der Index des Schubes, der zurückgegeben werden soll
   * @return null, falls "index" außerhalb der bereits bearbeiteten Schübe liegt,
   *         sonst ein BigInteger-Feld mit den Ausgabenachrichten des durch "index" bezeichneten
   *         Schubes
   */
  
  public BigInteger[] getOutBatch(int index){
    if((index >= 0) && (index < batchNumber)){
    	// clone the batch before outputting it 
    	BigInteger[] result = new BigInteger[BATCH_SIZE];
    	for(int i = 0; i < BATCH_SIZE; i++) {
    		result[i] = outBatches.elementAt(index)[i];
    	}
    	return result;    	
    }
    return null;
  }

  /**
   * Testet, ob ein Angriff korrekte Ergebnisse erzeugt hat. Für eine angegebene
   * Schubnummer wird ermittelt, ob die angegebene Nachrichtenzuordnung einer
   * Eingabenachricht zu einer Ausgabenachricht korrekt ist.
   * @param batch der Schub, in dem die Nachrichtenzuordnung gestestet werden soll
   * @param inIndex Index der Eingabenachricht im Feld der Eingabenachrichten des Schubes
   * @param outIndex Index der Ausgabenachricht im Feld der Ausgabenachrichten des Schubes
   * @return true, wenn die Zuordnung stimmt, sonst false
   * @exception Exception wenn indices außerhalb der Grenzen liegen: 0 <= batch < batchNumber,
   *            0 <= inIndex < BATCH_SIZE, 0 <= outIndex < BATCH_SIZE
   */
  
  public boolean checkAttackSuccess(int batch, int inIndex, int outIndex) throws Exception{
    if((batch < 0) || (batch >= batchNumber)
       || (inIndex < 0) || (inIndex >= BATCH_SIZE)
       || (outIndex < 0) || (outIndex >= BATCH_SIZE)){
      throw new Exception(Attack.messages.getString("MixFeldInd"));
    }
    // Eingabenachricht entschlüsseln
    // decrypt the input message
    BigInteger outMsg = RSA.decrypt(mixPrivateKey, ((BigInteger[])inBatches.elementAt(batch))[inIndex]);
    // entfernen der Zufallsbits
    // remove the random bits
    outMsg = outMsg.xor(outMsg.shiftRight(messageLength-randomLength)
                              .shiftLeft(messageLength-randomLength));
    if(outMsg.compareTo(((BigInteger[])outBatches.elementAt(batch))[outIndex]) == 0){
      return true;
    }
    return false;
  }

}
