
package algorithms;

import java.math.BigInteger;
import java.util.*;

/** 
 * <b>CRA - Algorithmus</b>
 * <P>
 * Diese Klasse implementiert den Chinesischen Restsatz Algorithmus.
 * <br>
 * <p> Verwendung:
 * <ol>
 *   <li> <b>Initialisierung:</b> <br>CRA myCRA = new CRA(reste[], moduln[]);
 *   <li> <b>Berechnung:</b> <br>myCRA.executeCRA();
 *   <li> <b>Ergebnis ermitteln:</b> <br>
 *        erg = myCRA.getErgebnis(); <br>
 *        gesamtModulus = myCRA.getMod();
 * </ol>  
 * @version 1.01, 99/11/03
 * @author  Sebastian Schulz, Sebastian Clauß
 * @since JDK1.2
 */


public class CRA {

    
    /* Fuer Internationalisierung */
    public static ResourceBundle messages; 
	
    /**
     * Ein Array von BigIntegern für die Reste des Modularen Gleichungssystems <p>
     *	@see java.math.BigInteger
     */
    
    private BigInteger[] r;
	
    /**
     * Ein Array von BigIntegern f&uuml;r die Modi des Modularen Gleichungssystems <P>
     *	@see java.math.BigInteger
     */
    
    private BigInteger[] m;
    /** M = m[0] * m[1] * ... * m[k] */
    private BigInteger M;
    /** Das Ergebnis x mod M */ 
    
    private BigInteger x;

	
    /**
     *	Instanziiert die CRA-Klasse mit allen erforderlichen Parametern.
     * <br>Das CRA-Objekt kann sofort veranlasst werden, sich zu berechnen.<P>
     * @see #executeCRA
     * @param reste die Reste
     * @param moduln die Moduln
     * @exception Exception wenn Resteanzahl != Modulanzahl, <br> 
     *                      keine Reste oder Moduln angegeben wurden, <br>
     *                      Moduln nicht paarweise Teilerfremd sind, <br>
     *                      Moduln nicht > 0 sind.
     */
    
    public CRA(BigInteger[] reste, BigInteger[] moduln) throws Exception{
	
	/* Der folgende Teil ist nötig für die Internationalisierung*/
	Locale currentLocale;
	currentLocale = Locale.getDefault();
	messages = ResourceBundle.getBundle("algorithmslocalmessages",currentLocale);


	setParams(reste, moduln);
    }
	
	
    /**
     *	Initialisiert die CRA-Klasse mit allen erforderlichen Parametern.
     * <br>Das CRA-Objekt kann sofort veranlasst werden, sich zu berechnen.<P>
     * @see #executeCRA
     * @param reste die Reste
     * @param moduln die Moduln
     * @exception Exception wenn Resteanzahl != Modulanzahl, <br> 
     *                      keine Reste oder Moduln angegeben wurden, <br>
     *                      Moduln nicht paarweise Teilerfremd sind, <br>
     *                      Moduln nicht > 0 sind.
     */
    
    public void setParams(BigInteger[] reste, BigInteger[] moduln) throws Exception{
	if (reste.length < 1 || moduln.length < 1) {
	    throw new Exception(messages.getString("CRAUngRestMod"));
	} else if (reste.length != moduln.length) {
	    throw new Exception(messages.getString("CRAAnzRestAnzMod"));
	} 
	r = reste;
	/* Moduln müssen paarweise Teilerfremd sein */
	
	for(int i = 0; i < moduln.length-1; i++){
	    for(int j = i+1; j < moduln.length; j++){
		if(moduln[i].gcd(moduln[j]).compareTo(BigInteger.valueOf(1)) != 0){
		    throw new Exception(messages.getString("CRAMod")+" "+Integer.toString(i)+" "+messages.getString("CRAUnd")+
					" "+ Integer.toString(j) +" "+messages.getString("CRANotTF"));
		}
	    }
	}
	/* Moduln müssen > 0 sein */
	
	for(int i = 0; i < moduln.length; i++){
	    if (moduln[i].compareTo(BigInteger.valueOf(0)) <= 0 ) {
		throw new Exception(messages.getString("CRAMod2")+" "+Integer.toString(i)+" "+messages.getString("CRANot0"));
	    }
	}
	m = moduln;
    }

    
    /**
     *	Gibt das aktuelle Array von Resten zurück.
     * @see #setParams
     * @return Reste
     */
    
    public BigInteger[] getReste(){
	return r;
    }

    
    /**
     *	Gibt das aktuelle Array der Modi zurück.
     * @see #setParams
     * @return Moduln
     */
     
    public BigInteger[] getModi(){
	return m;
    }
	
	
    /**
     *	Gibt die aktuelle Gesamtmodularität des modularen Gleichungssystems zurück.
     * @see #getErgebnis 
     * @return Modulus M = m[0] * m[1] * ... * m[k]
     */
    
    public BigInteger getMod(){
	return M;
    }

	
    /**
     *	Gibt das Gesamtergebnis des modularen Gleichungssystems zurück.
     * @see #getMod
     * @return Ergebnis x mod M
     */
    
    public BigInteger getErgebnis(){
	return x;
    }
	

    /**
     * Diese Methode berechnet das Ergebnis des CRA unter Verwendung der gesetzten Werte.
     * <br>(entweder durch den Konstruktor oder die 'setParams'-Methode) 
     * @see java.math.BigInteger
     * @see #setParams
     */
    
    
    public void executeCRA(){
	try{
	    BigInteger buffer;	                                     
	    /* Initialisierungen */
	    
	    M = m[0];               
	    x = r[0].mod(M);
	    
	    for(int i = 1; i < r.length; i++){
		/* für alle Paare (r,m) */
		
		// h = (r[i] - x)*M^-1 mod m[i]
		BigInteger h = M.modInverse(m[i]).multiply(r[i].subtract(x));
		
		x = x.add(h.multiply(M));                                
		// x = x + h*M
		
		M = M.multiply(m[i]);                                   
		// M = M*m[i]
		
	    }
	    
	    x = x.mod(M);
	    /* Ergebnis festlegen*/
	    
	}catch (Exception e){
	}
    }	
}
	
	
	
	
	
	
	
	
	
	
	
	
