/* Legendre-Symbol - Algorithmus */

//Sebastian Schulz 1999

package algorithms; 

import java.math.BigInteger;
import java.util.*;

/**
 * <b>Legendre-Symbol - Algorithmus</b>
 * <P>
 * Diese Klasse implementiert den Algorithmus zur Ermittlung des
 * Legendre-Symbols zum Test auf quadratische Reste.
  * <br>
  * <p> Verwendung:
  * <ol>
  *   <li> <b>Initialisierung:</b> <br>LEGENDRE myLEGENDRE = new LEGENDRE(testQuadRest, primzahl);
  *   <li> <b>Berechnung:</b> <br>myLEGENDRE.executeLEGENDRE();
  *   <li> <b>Ergebnis ermitteln:</b> <br>erg = myLEGENDRE.getErgebnis();
  * </ol>  
 * @version 1.01, 99/11/03
 * @author  Sebastian Schulz, Sebastian Clauß
 * @since JDK1.2
 */

public class LEGENDRE{
    
    
    /* Fuer Internationalisierung */
    public static ResourceBundle messages; 

    /** die Primzahl p */
    
    private BigInteger p;
    
    /** die Zahl, für die getestet werden soll, ob sie ein quadratischer Rest von p ist. */
    
    private BigInteger a;
    
    /** Das Legendre-Symbol L(a,p) */
    
    private int Ergebnis;
	

    /**
     *	Instanziiert die LEGENDRE-Klasse mit allen erforderlichen Parametern.
     * <br>Das LEGENDRE-Objekt kann mit Hilfe der Methode executeLEGENDRE sofort veranlasst 
     * werden, sich zu berechnen.<P>
     * @see #executeLEGENDRE
     * @param evtlQRest die zu testende Zahl a
     * @param primzahl der Modulus (Primzahl) p
     * @exception Exception wenn Primzahl p <= 2, <br>
     *                      p keine Primzahl ist.
     */
       
    public LEGENDRE(BigInteger evtlQRest, BigInteger primzahl)throws Exception{
	
	/* Der folgende Teil ist nötig für die Internationalisierung*/
	Locale currentLocale;
	currentLocale = Locale.getDefault();
	messages = ResourceBundle.getBundle("algorithmslocalmessages",currentLocale);


	setParams(evtlQRest, primzahl);
    }


    /**
     *	Initialisiert die LEGENDRE-Klasse mit allen erforderlichen Parametern.
     * <br>Das LEGENDRE-Objekt kann sofort veranlasst werden, sich zu berechnen.<P>
     * @see #executeLEGENDRE
     * @param evtlQRest die zu testende Zahl a
     * @param primzahl der Modulus (Primzahl)  p
     * @exception Exception wenn Primzahl p <= 2, <br>
     *                      p keine Primzahl ist.
     */
       
    public void setParams(BigInteger evtlQRest, BigInteger primzahl)throws Exception{
	if (primzahl.compareTo(BigInteger.valueOf(2)) <= 0){
	    throw new Exception(messages.getString("LEGENDREp2"));
	}
	if (!primzahl.isProbablePrime(200)){
	    throw new Exception(messages.getString("LEGENDREpPrime"));
	}
	a = evtlQRest;
	p = primzahl;
    }

    /**
     *	Gibt den aktuelle Wert des Test-Kandidaten zurück.
     * @see #setParams
     * @return die zu testende Zahl a
     */
    
    public BigInteger getA(){
	return a;
    }
	

    /**
     *	Gibt den aktuelle Wert der Primzahl zurück.
     * @see #setParams
     * @return den Modulus (Primzahl) p
     */
     
    public BigInteger getP(){
	return p;
    }

    /**
     * Gibt das Ergebnis zurück.
     * @see #setParams
     * @see #executeLEGENDRE
     * @return das Legendre-Symbol L(a,p)
     */
    
    public int getErgebnis(){
	return Ergebnis;
    }

	
    /**
     * Diese Methode berechnet das Ergebnis des Legendre-Algorithmus
     * unter Verwendung der gesetzten Werte.
     * <br>(entweder durch den Konstruktor oder die 'setParams'-Methode)
     * <br> Das Ergebnis wird mit der Methode getErgebnis zurückgegeben.
     * @see #setParams
     * @see #getErgebnis
     */
    
    public void executeLEGENDRE(){
	try{                                           
	    //exp = (p - 1)/2  
	    BigInteger exp = p.subtract(BigInteger.valueOf(1)).divide(BigInteger.valueOf(2));
	    
	    BigInteger erg = a.modPow(exp,p);   
	    // Legendre symbol(a,p) = a^exp mod p

	    /* betragskleinsten Rückgabewert erzeugen */
	    
	    if (erg.compareTo(BigInteger.valueOf(1)) > 0){  
 		Ergebnis = -1;                        
		/* modPow gibt positive Werte zurück */
		
	    }else{                                    
		/* muß in betragskleinsten Wert gewandelt werden. */
		
		Ergebnis = erg.intValue();
	    }
	} 
	catch (Exception e){
	    System.out.println("LEGENDRE.executeLEGENDRE: "+e.getMessage());
	    e.printStackTrace();
	}
    }
}
	
	
	
	
	
	
	
	
	
	
	
	
