/* SMA - Algorithmus */

//Sebastian Schulz 1999

package algorithms;

import java.math.BigInteger;
import java.util.*;

/** 
 * <b>SQRTMOD - Algorithmus</b>
 * <P>
 * Diese Klasse implementiert den SQRTMOD - Algorithmus. Mit diesem Algorithmus können die Wurzeln
 * aus einem quadratischen Rest eines Produktes zweier Primzahlen p*q gezogen werden.
 * <br>
 * <p> Verwendung:
 * <ol>
 *   <li> <b>Initialisierung:</b> <br>SQRTMOD mySQRTMOD = new SQRTMOD(primP. primQ, quadRest);
 *   <li> <b>Berechnung:</b>  <br>mySQRTMOD.executeSQRTMOD();
 *   <li> <b>Ergebnis ermitteln:</b> <br>erg[] = mySQRTMOD.getErgebnis();
 * </ol>  
 * @version 1.01, 99/11/03
 * @author  Sebastian Schulz, Sebastian Clauß
 * @since JDK1.2
 */

public class SQRTMOD{
    
    /* Fuer Internationalisierung */
    public static ResourceBundle messages; 
    /** die Primzahl p = 3 mod 4 */
    
    private BigInteger p;
    /** die Primzahl q = 3 mod 4 */
    
    private BigInteger q;
    /** die Zahl a (QR(p*q)) aus der die Wurzel gezogen werden soll. */
    
    private BigInteger a;
    /** die quadratischen Reste */
    
    private BigInteger[] r = new BigInteger[2];

    /**
     *	Instanziiert die SQRTMOD-Klasse mit allen erforderlichen Parametern.
     * <br>Das SQRTMOD-Objekt kann mit Hilfe der Methode executeSQRTMOD sofort veranlasst werden, 
     * sich zu berechnen.<P>
     * @see #executeSQRTMOD
     * @param primP Primzahl p = 3 mod 4
     * @param primQ Primzahl q = 3 mod 4
     * @param qrA   quadratischer Rest a (von p*q)
     * @exception Exception wenn p oder q keine Primzahl ist, <br>
     *                      p = 3 mod 4 oder q = 3 mod 4 nit erfüllt ist, <br>
     *                      a kein quadratischer Rest von p*q ist.
     */
    
    public SQRTMOD(BigInteger primP, BigInteger primQ, BigInteger qrA)throws Exception{
	
	/* Der folgende Teil ist nötig für die Internationalisierung */
	Locale currentLocale;
	currentLocale = Locale.getDefault();
	messages = ResourceBundle.getBundle("algorithmslocalmessages",currentLocale);


	setParams(primP, primQ, qrA);
    }


    /**
     *	Initialisiertst die SQRTMOD-Klasse mit allen erforderlichen Parametern.
     * <br>Das SQRTMOD-Objekt kann sofort veranlasst werden, sich zu berechnen.<P>
     * @see #executeSQRTMOD
     * @param primP Primzahl p = 3 mod 4
     * @param primQ Primzahl q = 3 mod 4
     * @param qrA   quadratischer Rest a (von p*q)
     * @exception Exception wenn p oder q keine Primzahl ist, <br>
     *                      p = 3 mod 4 oder q = 3 mod 4 nit erfüllt ist, <br>
     *                      a kein quadratischer Rest von p*q ist.
     */
        
    public void setParams(BigInteger primP, BigInteger primQ, BigInteger qrA)throws Exception{
	if(!primP.isProbablePrime(200)){
	    throw new Exception(messages.getString("SQRTMODpPrime"));
	}else if (primP.mod(BigInteger.valueOf(4)).compareTo(BigInteger.valueOf(3)) != 0){
	    throw new Exception(messages.getString("SQRTMODp3Mod4"));
	}else if (primP.compareTo(BigInteger.valueOf(2)) <= 0){
	    throw new Exception(messages.getString("SQRTMODp2"));
	}
	p = primP;

	if(!primQ.isProbablePrime(200)){
	    throw new Exception(messages.getString("SQRTMODqPrime"));
	}else if (primQ.mod(BigInteger.valueOf(4)).compareTo(BigInteger.valueOf(3)) != 0){
	    throw new Exception(messages.getString("SQRTMODq3Mod4"));
	}else if (primQ.compareTo(BigInteger.valueOf(2)) <= 0){
	    throw new Exception(messages.getString("SQRTMODq2"));
	}
	q = primQ;

	LEGENDRE l1 = new LEGENDRE(qrA,p);
	l1.executeLEGENDRE();
	LEGENDRE l2 = new LEGENDRE(qrA,q);
	l2.executeLEGENDRE();
	if ((l1.getErgebnis() != 1) || (l2.getErgebnis() != 1)){
	    throw new Exception(messages.getString("SQRTMODquadrRest"));
	}
	a = qrA;
    }


    /**
     *	Gibt den aktuelle Wert des Parameters P zurück.
     * @see #setParams
     * @return Primzahl p
     */
    
    public BigInteger getP(){
	return p;
    }
	

    /**
     *	Gibt den aktuelle Wert des Parameters Q zurück.
     * @see #setParams
     * @return Primzahl q
     */   
    
    public BigInteger getQ(){
	return q;
    }
	

    /**
     *	Gibt den aktuelle Wert des Parameters A zurück.
     * @see #setParams
     * @return der quadratische Rest a
     */
    
    public BigInteger getA(){
	return a;
    }


    /**
     *	Gibt den aktuelle Wert des Modulos (P*Q) zurück.
     * @see #setParams
     * @return m = p*q
     */
    
    public BigInteger getM(){
	return p.multiply(q);
    }


    /**
     *	Gibt die Lösungen zurück.
     * @see #executeSQRTMOD
     * @return r ein Feld mit allen gefundenen Lösungen
     */
    
    public BigInteger[] getErgebnis(){
	return r;
    }


    /**
     * Diese Methode berechnet das Ergebnis des SQRTMOD unter Verwendung der gesetzten Werte.
     * <br>(entweder durch den Konstruktor oder die 'setParams'-Methode)
     * gegeben sind die Primzahlen p und q und der quadratische Rest a.
     * @see java.math.BigInteger
     * @see CRA
     * @see #getErgebnis
     */
       
    public void executeSQRTMOD(){
	try{
	    BigInteger m = p.multiply(q);                      // m = p * q
	    
	    BigInteger exp = p.add(BigInteger.valueOf(1)).divide(BigInteger.valueOf(4));	
	    BigInteger rp = a.modPow(exp,p);                   // rp = a^((p+1)/4) mod p
	    
	    exp = (q.add(BigInteger.valueOf(1))).divide(BigInteger.valueOf(4));	
	    BigInteger rq = a.modPow(exp,q);                  // rq = a^((q+1)/4) mod q
	    
	    BigInteger reste[]  = {rp, rq};
	    BigInteger moduln[] = {p,q};
	    CRA cra = new CRA(reste,moduln);
	    cra.executeCRA();
	    r[0] = cra.getErgebnis();                         // r[0] = CRA(rp,rq, p, q)
	    
	    reste[0] = rp;
	    reste[1] = rq.negate();
	    cra = new CRA(reste,moduln);
	    cra.executeCRA();
	    r[1] =  cra.getErgebnis();                        // r[1] = CRA(rp, -rq, p, q)
	}
	catch (Exception e){
	}
    }

}
	
	
	
	
	
	
